export const e = 1e-6;

export class Vector {
  x: number;
  y: number;

  constructor(x = 0, y = 0) {
    this.x = x;
    this.y = y;
  }

  add(vec: Vector) {
    return new Vector(vec.x + this.x, vec.y + this.y);
  }

  sub(vec: Vector) {
    return new Vector(this.x - vec.x, this.y - vec.y);
  }

  div(scalar: number) {
    return new Vector(this.x / scalar, this.y / scalar);
  }

  mul(scalar: number) {
    return new Vector(this.x * scalar, this.y * scalar);
  }

  len() {
    return Math.sqrt(this.x * this.x + this.y * this.y);
  }

  norm() {
    if (this.len() < e) return new Vector(1, 0);
    return this.div(this.len());
  }

  dotProd(vec: Vector) {
    return new Vector(this.x * vec.x, this.y * vec.y);
  }

  det(vec: Vector) {
    return this.x * vec.y - this.y * vec.x;
  }

  turnRight() {
    return new Vector(this.y, -this.x);
  }

  turnLeft() {
    return new Vector(-this.y, this.x);
  }

  closestTo(vec: Vector) {
    if (this.len() <= e) return vec;

    const det = this.det(vec);
    const closestVec = det > 0 ? this.turnRight() : this.turnLeft();
    return closestVec.mul(Math.abs(det) / Math.pow(this.len(), 2));
  }

  distToSegment(vec: Vector) {
    const closestVec = this.closestTo(vec);
    const intersection = vec.add(closestVec.mul(-1));

    const distA = intersection.len();
    const distB = intersection.sub(this).len();

    if (distA > this.len() || distB > this.len()) {
      if (distA > distB) {
        return vec.sub(this).len();
      } else {
        return vec.len();
      }
    } else {
      return closestVec.len();
    }
  }
}

export class MassPoint {
  position: Vector;
  momentum: Vector;
  force: Vector;
  dumping: number;

  constructor(vec: Vector, dumping: number) {
    this.position = vec;
    this.momentum = new Vector(0, 0);
    this.force = new Vector(0, 0);
    this.dumping = dumping;
  }

  applyForce(vec: Vector) {
    if (isNaN(vec.x) || isNaN(vec.y)) {
      throw new Error("found NaN applyForce");
    }
    this.force = this.force.add(vec);
  }

  update(time: number) {
    const acc = this.force;
    this.force = new Vector(0, 0);

    this.momentum = this.momentum
      .mul(Math.min(1, Math.max(0, Math.pow(1 - this.dumping, time * 10))))
      .add(acc.mul(time));
    this.position = this.position.add(this.momentum.mul(time));
  }
}

const w = window as any;
w.Vector = Vector;
w.MassPoint = MassPoint;
